--------------------------------------------------------------------
--!@file Mudules used in simulation of Newton Raphson method 
--------------------------------------------------------------------

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.numeric_std.ALL;



package NewtonRaphsonSimComponents is
	--!@brief normalization of a std_logic_vector.
	--!@param vec the std_logic_vector to normalize
	--!@param shift the number of position shifted 	 
	--!@note if vec = "00..0" then shift := vec'high
	procedure norm(	variable vec : inout std_logic_vector; variable shift : out integer);
	
	--!@brief calculates initial aproximation of the resiprocal
	--!@param input high part of the divisor
	--!@param size bit width of the input and output
	--!@details the returned value is in range [0.5, 1.0> first bit is 2^-1
	function table_loockup(input : std_logic_vector; size : natural) return std_logic_vector;
	
	--!@brief calculates one iteration of the newton raphson algorithm
	--!@param b the divisor input, internaly truncated to appropriate width
	--!@param r the input approximation of the resiprocal
	--!@return the new inproved approximation of the resiprocal
	function NewtonRaphsonStage(b, r : std_logic_vector)
		return std_logic_vector;
			
	function Max(a, b : integer)return integer;
	function Min(a, b : integer)return integer;
	
end NewtonRaphsonSimComponents;

package body NewtonRaphsonSimComponents is
	
	procedure norm(variable vec : inout std_logic_vector; variable shift : out integer)is
		variable index : integer := 0;
		variable retur : std_logic_vector(vec'range);
	begin
	   
		retur := vec;
		while retur(retur'high) /= '1' and index < retur'high loop
			retur := retur(retur'high - 1 downto 0) & '0';
			index := index + 1;
		end loop;
		vec := retur;
		shift := index;
	end norm;
	
	function table_loockup(input : std_logic_vector; size : natural) return std_logic_vector is
		variable retur : std_logic_vector(size downto 0);
		variable a, b, c : real;
	begin
		a := (1.0)/real(2**(input'length - 1));
		a := a * real(to_integer(IEEE.numeric_std.unsigned(input)));--1.0;
		a :=  (1.0) / a;
		a := a * real(2**(size + 1));
		retur := conv_std_logic_vector(integer(a), size + 1);
		
		--if the truncated representation of b = 1,00...0 the retur value will overflow, so we round one down.
		if retur = (retur'range => '0') then
			retur := (others => '1');
		end if; 
		
		return retur(retur'high downto 1);
	end table_loockup;
	

	--Newton-Raphson simulation stage
	function NewtonRaphsonStage(b, r : std_logic_vector)
		return std_logic_vector is
		--{nr of integer nr : number of decimals}

		--!output precision
		constant k : integer := r'length * 2;

		--!truncated representation of b
		variable b_trunc : std_logic_vector(min(b'high, k+2) downto 0);--{1:x}
		--!result after firts multiplication
		variable mult_one : std_logic_vector(b_trunc'length + r'length - 1 downto 0);--{1:95}
		--!truncated representation of first multiplication
		variable mult_one_trunc : std_logic_vector(k+2 downto 0);--{1:k+2}
		--!resoult after second multiplication
		variable second_mult : std_logic_vector(mult_one_trunc'length + r'length - 1 downto 0);--{1:k+2 + r'lengt}
		--!truncated representation of second multiplication
		variable new_r :  std_logic_vector(k-1 downto 0);

	begin
	
		b_trunc := b(b'high downto b'length - b_trunc'length);
		mult_one := b_trunc * r;
		mult_one_trunc := mult_one(mult_one'high downto mult_one'length - mult_one_trunc'length);

		mult_one_trunc := not(mult_one_trunc) + 1;

		second_mult := mult_one_trunc * r; --range <0.0,1.0> format{1:xxxx}
		new_r := second_mult(second_mult'high - 1 downto second_mult'length - new_r'length - 1);

		return new_r;
		
	end NewtonRaphsonStage;
	
	
	function Max(a, b : integer)return integer is
	begin
		if a > b then
			return a;
		else
			return b;
		end if;
	end Max;
	
	function Min(a, b : integer)return integer is
	begin
		if a > b then
			return b;
		else
			return a;
		end if;
	end Min;	
	
end NewtonRaphsonSimComponents;
