---------------------------------------------------------
--(copy left) Arthur Øvergaard
--Trondheim June 16, 2009
--
--Implements Newton-Raphson simulation test bench
---------------------------------------------------------

library ieee, work;

use ieee.std_logic_1164.all;
use IEEE.numeric_std.ALL;
use work.txt_util.all;

entity NewtonSim_tb is end NewtonSim_tb;

architecture behavior of NewtonSim_tb is

	component sim_newton is 
		port(   dividend :	in std_logic_vector(63 downto 0);
			divisor : 		in std_logic_vector(63 downto 0);
			clock : 		in std_logic;
			quotient : 		out std_logic_vector(63 downto 0);
			rest : 			out std_logic_vector(63 downto 0));                      
	end component;

	constant period : time := 10 ns;

	signal tb_dividend : std_logic_vector(63 downto 0);
	signal tb_divisor : std_logic_vector(63 downto 0);
	signal tb_clock : std_logic := '0';
	signal tb_quotient: std_logic_vector(63 downto 0);
	signal tb_rest : std_logic_vector(63 downto 0);
	
begin

	--Instantiation of simulation model
	st:sim_newton port map(
			tb_dividend,
			tb_divisor,
			tb_clock,
			tb_quotient,
			tb_rest);


	--Clocking process
	clocking:process 
	begin
		wait for period/2;
		tb_clock <= not( tb_clock );
	end process;
	
	--Generate test vectors
	process
		variable divd, divs : std_logic_vector(tb_dividend'range);
		variable pattern : std_logic_vector(tb_dividend'range);
	begin
	
		
		
		--Generate manual test vector1
		divd := (others => '0');
		divd(divd'high) := '1';

		divs := (others => '0');
		divs(divs'high) := '1';
		divs(divs'high-3) := '1';
				
		tb_dividend <= divd;
		tb_divisor  <= divs;
		
		wait until tb_clock'event and tb_clock = '1';

		--Generate manual test vector2
		divs(divs'high-4) := '1';
		
		tb_dividend <= divd;
		tb_divisor  <= divs;
		
		wait until tb_clock'event and tb_clock = '1';
		

		--Generate patterns of waling one's on dividend and divisor 
		for i in 0 to 63 loop
			divd := (others => '0');
			divd(i) := '1';
			for j in 0 to 63 loop
				divs := (others => '0');
				divs(j) := '1';

				tb_dividend <= divd;
				tb_divisor  <= divs;
				
				wait until tb_clock'event and tb_clock = '1';

			end loop;
		end loop;

		--Generate patterns of walking one's on the dividend with high bit '1'
		--Generate increasingly larger patterns of one's to apply to the divisor
		for i in 0 to 63 loop
			divd := (others=>'0');
			divd(divd'high) := '1';
			divd(i) := '1';
			
			for p_len in 1 to 62 loop--for pattern lengths from 1 to 62
				
				--generate pattern of one's
				pattern := (others => '0');
				for j in 0 to p_len - 1 loop pattern(j) := '1'; end loop;

				--for all shifted positions of the pattern
				for j in 0 to 64 - p_len loop
					divs := (divs'high => '1', others => '0');
					divs := divs or pattern;
					pattern := pattern(pattern'high - 1 downto 0) & '0';

					tb_dividend <= divd;
					tb_divisor  <= divs;
					     
					wait until tb_clock'event and tb_clock = '1';
					
				end loop;
			end loop;
		end loop;

		--Generate patterns of walking one's on the dividend with high bit '1'
		--Generate increasingly larger patterns of zero's to apply to the divisor
		for i in 0 to 63 loop
			divd := (others=>'0');
			divd(divd'high) := '1';
			divd(i) := '1';
			for p_len in 1 to 62 loop

				--generate pattern of zero's
				pattern := (others => '1');
				for j in 0 to p_len - 1 loop pattern(j) := '0'; end loop;

				for j in 0 to 64 - p_len loop
					divs := (others => '1');
					divs := divs and pattern;
					pattern := pattern(pattern'high - 1 downto 0) & '1';

					tb_dividend <= divd;
					tb_divisor  <= divs;

					wait until tb_clock'event and tb_clock = '1';

				end loop;
			end loop;
		end loop;       
		report "Simulation finished";
		wait;
	end process;

	--verify the answer
	verify:process
		variable q, r : std_logic_vector(63 downto 0);
		variable s, d : string(1 to 64);--strings for text printing
	begin

		wait until tb_clock'event and tb_clock = '1';

		q := std_logic_vector(unsigned(tb_dividend) / unsigned(tb_divisor));--Calculate the exact quotient
		r := std_logic_vector(unsigned(tb_dividend) rem unsigned(tb_divisor));--Calculate the exact reminder    

      wait until tb_clock = '0';
      
		if tb_quotient /= q or tb_rest /= r then --Compere result
			s := str(tb_divisor);
			d := str(tb_dividend);
			report "Error: " & s & " : " & d;

			s := str(tb_quotient);
			d := str(q);
			report "Sim is: " & s & ", ans: " & d;

		end if;

	end process;


end behavior;

