#include <math.h>
#include <iostream>
#include <fstream>
#include "Decomposition.h"
#include "Exception.h"

#define MAX(a, b) ((a)>(b) ? (a) : (b))
#define MIN(a, b) ((a)<(b) ? (a) : (b))
/**
 *Constructor assigns
 */
Decomposition::Decomposition(int Wi_, int Wo_, int TOs, int alpha_, const int* gamma_, const int * beta_)
{
	int p_accumulate = 0;
	long double errorLimit;
	mTOs = TOs;
	
	Wi = Wi_;
	Wo = Wo_;
	alpha = alpha_;
	
	gamma = new int[mTOs];
	beta = new int[mTOs];
	p = new int[mTOs];
	
	for(int i=0;i<mTOs;i++){
		gamma[i] = gamma_[i];
		beta[i] = beta_[i];
		p[i] = p_accumulate;
		p_accumulate += beta[i];
	}
	
	ApproxError = calcMathApproxError();
	errorLimit = 0.5 * pow(2, -((long double)(Wo + 1)));
	if(ApproxError < errorLimit)errorOK = true;
	else errorOK = false;
}

/**
 *Copy constructor allocates nessesary memory and copys values
 @param other class to be copyed
 */
Decomposition::Decomposition(const Decomposition& other)
{
	Wi = other.Wi;
	Wo = other.Wo;
	alpha = other.alpha;
	mTOs = other.mTOs;
	
	ApproxError = other.ApproxError;
	errorOK = other.errorOK;

	gamma = new int [mTOs];
	beta = new int [mTOs];
	p = new int [mTOs];
	
	for(int i=0;i<mTOs;i++){
		gamma[i] = other.gamma[i];
		beta[i] = other.beta[i];
		p[i] = other.p[i];
	}
}

/**
 *Loade one decomposition from file
 @note ifstream pointer must point to the start of an decomposition
 */
Decomposition::Decomposition(std::fstream& file){

	if(!file.is_open())throw Exception("File not open error");;
	
	file.read((char*)&mTOs, sizeof(int));
	file.read((char*)&alpha, sizeof(int));
	file.read((char*)&Wi, sizeof(int));
	file.read((char*)&Wo, sizeof(int));
	file.read((char*)&ApproxError, sizeof(long double));
	
	gamma = new int[mTOs];
	beta = new int[mTOs];
	p = new int[mTOs];
	
	for(int i=0;i<mTOs;i++){
		file.read((char*)&gamma[i], sizeof(int));
		file.read((char*)&beta[i], sizeof(int));
		file.read((char*)&p[i], sizeof(int));
	}
}

/**
 *Calculate the 'infinite' accurate x value at Ci for TO table i
 @param Ci is the value x in f(x)
 @param i is the TO number
*/
long double Decomposition::Xleft(unsigned int Ci, int i)const
{
	long double temp;

	if(i >= mTOs || i < 0)throw Exception("TO range error in Xleft()");
	
	temp = pow(2,-(long double)gamma[i]);
	temp = temp * ((long double)( Ci ));
	temp += 1;
	return temp;
}

/**
 *Calculate the 'infinite' accurate x value at (Ci + 1) - TIVstep
 @param Ci is the value x in f(x)
 @param i is the TO number
 */
long double Decomposition::Xright(unsigned int Ci, int i)const
{
	long double power1, power2, ret;
	int power;
	
	if(i >= mTOs || i < 0)throw Exception("TO range error in Xright()");
	
	power = Wi - p[i] - beta[i];
	
	power1 = pow(2, -(long double)gamma[i]); 
	power2 = pow(2, -(long double)power);

	ret = ((power1 * (long double)(Ci + 1)) - power2);
	ret += 1;

	return ret;
}

/**
 *Calculates the range  along x of TO table i
 @param i is the TO number
 */
long double Decomposition::Delta(int i)const
{
	long double temp;

	if(i >= mTOs || i < 0)throw Exception("TO range error in Delta()");
	
	temp = pow(2, (long double)(-Wi + p[i])) * ((long double)( (1<<beta[i]) - 1));
	return temp;
}

/**
 *Destructor
 @todo delete of TIV and TOs
 */
Decomposition::~Decomposition()
{
	delete [] gamma;
	delete [] beta;
	delete [] p;
}

/**
 *Calculates the maximum error this decomposition will have due to mathematical approximation
 */
long double Decomposition::calcMathApproxError()const
{
	long double x1, x2, x3, x4, delta;
	long double eLow, eHigh;
	
	long double error = 0.0;
		
	for(int i = 0;i<mTOs;i++)
	{
		delta = Delta(i);
		//Calculate lowside error
		x1 = 1; 	//actualy 1 + (pow(2, -gamma[i]) * 0) wich is 1
		x2 = x1 + delta;
		x3 = x1 + pow(2,-(long double)gamma[i]) - pow(2, -( (long double) (Wi - p[i] - beta[i]) ));
		x4 = x3 + delta;
		
		eLow = (f(x2) - f(x1) - f(x4) + f(x3)) / 4;
		
		//Calculate highside error
		x1 = 1 + (pow(2, -(long double)gamma[i]) * ((1<<gamma[i]) - 1));
		x2 = x1 + delta;
		x3 = x1 + pow(2,-(long double)gamma[i]) - pow(2, -( (long double) (Wi - p[i] - beta[i]) ));
		x4 = x3 + delta;
		
		eHigh = (f(x2) - f(x1) - f(x4) + f(x3)) / 4;
		
		error += MAX(fabs(eLow), fabs(eHigh));
	}
	
	return error;
}

/**
 *Calculate number of entrys in the TIV and TOs
 @note this is based on the use of symetry in the TOs
 */
int Decomposition::Entrys(bool symmetry = true)const
{
	int ent = 1;
	int sym = 0;
	
	if(!errorOK)return 0x7FFFFFFF;//asumes 32bit int
	
	if(symmetry)sym = 1;

	ent = ent<<alpha;
	for(int i=0;i<mTOs;i++){
		ent += (1<<(gamma[i]+beta[i]-sym));//-1 due to symetry
	}
	return ent;
}

/**
 *Calculate number of guard bits needed to avoide rounding error when suming up the TIV and TOs
 */
int Decomposition::getGuardBits(void)const
{
	long double temp, frac;
	int ret;
	
	if(!errorOK)throw Exception("Error to greate");
	
	temp = 0.5 * (long double)mTOs;
	frac = (0.5 * pow(2, - ((long double)(Wo + 1)) ));
	temp =  temp / (frac-ApproxError);	
	//temp =  temp / frac; mod mulig feil implementert i første utkast
	temp = log(temp)/log(2); //log2(temp);
	//temp = (-(long double)Wo - 1 + temp);
	
	ret = (int)ceil(temp);
	
	return (-Wo - 1 + ret);
}

/**
 *Print this decomposition
 */
void Decomposition::Print(bool symmetry)const
{
	using namespace std;
	
	int sym = 0;
	if(symmetry)sym = 1;
	
	cout << "Set: A:" << alpha << " B:" << Wi - alpha;
	
	for(int i=mTOs - 1; i >= 0;i--){
		cout << " | Ai(" << i << "):" << gamma[i];
		cout << " Bi(" << i << "):" << beta[i];
	}
	
	cout << " | Error: " << ApproxError;
	
	if(ApproxError >= 0.5 * pow(2, -((long double)(Wo + 1))))std::cout << " *** ";
	cout << endl;

	int totSize, tempSize;
	
	totSize = (1<<alpha) * (Wo + getGuardBits());
	tempSize = totSize;
	
	cout << "  Table size  " << endl;
	cout << "  TIV:" << tempSize << " (" << (1<<alpha) << "*" << Wo + getGuardBits() << ")" << endl;
	for(int i=mTOs - 1;i >= 0;i--){
		tempSize = (1<<(gamma[i]+beta[i]-1)) * (getWidthTO(i) - 1);			
		totSize += tempSize;
		cout << "  TO(" << i << "):" << tempSize << " (" << (1<<(gamma[i]+beta[i]-sym)) << "*" << (getWidthTO(i) - sym) << ")"<< endl;
	}
	cout << "  ** TotalSize ** " << totSize << endl;

	cout << endl;
}

/**
 *Calculate the average slope in the range of Ci for TOi
 */
long double Decomposition::Slope(int Ci, int i)const
{	
	long double x1, x2, x3, x4, delta;
	
	if(i >= mTOs || i < 0)throw Exception("TO range error in Slope()");
	
	delta = Delta(i);
	
	x1 = 1 + (pow(2, -(long double)gamma[i]) * (long double)(Ci));
	x2 = x1 + delta;
	x3 = x1 + pow(2,-(long double)gamma[i]) - pow(2, -( (long double) (Wi - p[i] - beta[i]) ));
	x4 = x3 + delta;

	
	return ((f(x2) + f(x4)) - (f(x1) + f(x3))) / (2 * delta);
}

/**
 *Calculates the range of the TOs
 */
long double Decomposition::getRangeTO(int i)const
{
	long double delta, r1, r2;

	if(i >= mTOs || i < 0)throw Exception("TO range error in getRangeTO()");

	delta = Delta(i);
	r1 = Slope(0 ,i) * delta;
	r2 = Slope( ((1<<gamma[i])-1), i) * delta;
		
	return MAX(fabsl(r1), fabsl(r2));	
	
}

/**
 *Calculate the output width of the TOs
 */
int Decomposition::getWidthTO(int i)const
{
	int temp;
	
	if(i >= mTOs || i < 0)throw Exception("TO range error in WidthTO()");
	
	temp = (int)ceill( log( getRangeTO(i) / 0.5 ) / log(2) );
	
	temp += Wo + getGuardBits();
	
	return temp;
}

/**
 *Calculate the total size of this decomposition
 */
int Decomposition::CalcSize(bool symmetry = true)const
{
	int totSize, sym  = 0;

	if(symmetry)sym = 1;
	
	totSize = (1<<alpha) * (Wo + getGuardBits());
	
	for(int i=0;i<mTOs;i++){
		totSize += (1<<(gamma[i]+beta[i]-sym)) * (getWidthTO(i) - sym);	
	}
	return totSize;
}

/**
 *Save this decompositon to file
 */
void Decomposition::SaveToFile(std::fstream& fileRef)const
{
	
	if(!fileRef.is_open())throw Exception("File write error");
	
	fileRef.write((char*)&mTOs, sizeof(int));
	fileRef.write((char*)&alpha, sizeof(int));
	fileRef.write((char*)&Wi, sizeof(int));
	fileRef.write((char*)&Wo, sizeof(int));
	fileRef.write((char*)&ApproxError, sizeof(long double));


	
	for(int i=0;i<mTOs;i++){
		fileRef.write((char*)&gamma[i], sizeof(int));
		fileRef.write((char*)&beta[i], sizeof(int));
		fileRef.write((char*)&p[i], sizeof(int));
	}
}

void Decomposition::getTOparam(int i, int& gamma_, int &beta_, int& p_)const
{
	if(i >= mTOs || i < 0)throw Exception("TO range error in getTOparam()");
	
	gamma_ = gamma[i];
	beta_ = beta[i];
	p_ = p[i];
}

void Decomposition::getTIV(int* TIVbuffer, int & max_size)const
{
	long double xleft, xright, val;
	
	int k = getGuardBits();
	//int oddEven = mTOs % 2;
	int buffer_size = max_size;
	
	max_size = (1<<alpha);
	
	if(max_size > buffer_size)throw Exception("TIVbuffer smaler than TIV table size in getTIV()");
	
	for(int A = 0;A < buffer_size && A < (1<<alpha) ; A++)
	{
		xleft = 1 + pow(2, -(long double)(alpha)) * A;
		xright = xleft;
		
		for(int i = 0; i<mTOs;i++){
			xright += Delta(i);
		}		
		
		//val = ( f(xleft) + f(xright) ) / 2;
		
		//std::cout << "L:" << xleft << " R:" << xright << std::endl;
		//std::cout << "f(l)=" << f(xleft) << " f(r)=" << f(xright) << " avg=" << val << std::endl;
		
		val = ( f(xleft) + f(xright) ) / 2;
		
		TIVbuffer[A] = (int)floor(( (long double)(1<<(Wo + k)) * (val-0.5)*2 ));
		//std::cout << "tabval:" << std::hex << TIVbuffer[A] << std::dec << std::endl;
		TIVbuffer[A] +=  (mTOs/2) + (1<<(k - 1));
		
		//std::cout << "tabval:" << std::hex << TIVbuffer[A] << std::dec << std::endl;
	}
}

void Decomposition::getTO(int i, int* TObuffer, int & max_size, bool symmetry = true)const
{
	long double val, slope;
	int buffer_size = max_size;
	int sym = 0;
	int k = getGuardBits();
	
	if(symmetry)sym = 1;
	max_size = (1<<(gamma[i] + beta[i] - sym));
	
	if(i >= mTOs)throw Exception("TO range error in getTO()");
	if(max_size > buffer_size)throw Exception("TObuffer smaler than TO table size in getTO()");
		

	for(int C = 0;C < (1<<(gamma[i])) ; C++){
		slope = Slope(C, i);
		//std::cout << "Slope: " << slope << std::endl;
		for(int B = 0; B < (1<<beta[i] - sym) ; B++){
			val = slope * pow(2, (long double)(-Wi + p[i])) * ((long double)B + 0.5);
			
			TObuffer[(C * (1<<beta[i] - sym)) + B] = (int)floor((long double)(1<<(Wo + k + 1)) * val);
				
		}
	}
}

void Decomposition::Verify(bool verbose)const
{
	using namespace std;
	
	int alpha_mask;
	int gamma_mask[mTOs];
	int beta_mask[mTOs];
	int sym_mask[mTOs];
	
	int TIVdata[1<<alpha];
	int * TOdata[mTOs];
	unsigned int result, target, actual;
	int size, guard, TIVindex;
	int emax = 0, emin = 0;
	
	bool ok = true;
	
	guard = getGuardBits();
	
	alpha_mask = ((1<<alpha)-1) << (Wi-alpha);
	size = (1<<alpha);
	getTIV(TIVdata, size);
	
	for(int i = 0;i<mTOs;i++){
		size = (1<<(gamma[i] + beta[i] -1));
		gamma_mask[i] = ((1<<gamma[i])-1) << (Wi- gamma[i]);
		beta_mask[i] = ((1<<(beta[i]-1))-1) << p[i];
		sym_mask[i] = (1<<(beta[i] + p[i] -1));
		TOdata[i] = new int[size];
		getTO(i, TOdata[i], size, true);
	}
	
	for(int i = 1;i < (1<<Wi);i++){

		TIVindex = (i & alpha_mask)>> (Wi-alpha);
		result = TIVdata[TIVindex] | (1<<(Wi + guard));
		
		for(int j=0;j<mTOs;j++){
			unsigned int index, temp;
			index = (i & gamma_mask[j])>>(Wi - (gamma[j]+beta[j]-1));
			if(i & sym_mask[j]){
				index += ((i & beta_mask[j])>>p[j]);
				temp = (TOdata[j][index]);
			}else{
				index += ((~i & beta_mask[j])>>p[j]);
				temp = ~(unsigned int)(TOdata[j][index]);
			}
			result += temp;
		}
		
		target = (int)(powl(2.0,(long double)(2*Wo)) / (powl(2.0,(long double)Wo) + (long double) i));//org
		//target = (int)((powl(2.0,(long double)(2*Wo)) / (powl(2.0,(long double)Wo) + (long double) i))*(1<<guard));
		
		//target = ((target+1)>>1);
		//target = target << guard+1;
		
		actual = (unsigned int)((result>>(guard+1))&((1<<Wo)-1));//org
		//actual = (unsigned int)((result)&( ( 1<<(Wo+guard) )-1) );
		
		
		if(actual != target)ok = false;
		
		emin = MIN(((int)actual - (int)target), emin);
		emax = MAX(((int)actual - (int)target), emax);
		
		if(verbose){
			cout << hex << i << ":";
			cout << target << ":" << actual <<":";	
			cout << dec << (int)(actual - target) << hex << " " << TIVindex;
			
			for(int j=mTOs-1;j>=0;j--){
				if(i & sym_mask[j])cout << "+";
				else cout << "-";
			}
			
			cout << dec << endl;
		}
	}

	if(ok)cout << "Verify OK" << endl;
	else cout << "Verify FEILED!! max error:" << emax << " min error:" << emin << endl;
}
