------------------------------------------------------
--(copy left) Arthur Øvergaard 
--Trondheim June 09
--
--Implements all Newton-Rapson stages
------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.numeric_std.ALL;

library work;
use work.NR_implement_modules.NR_stage;
use work.NR_implement_modules.NR_stage_nopipe;

entity NR_stages is
	port(
		dividend_in :		in std_logic_vector(63 downto 0);
		divisor_in :		in std_logic_vector(63 downto 0);
		divisor_norm_in :	in std_logic_vector(63 downto 0);
		resiprocal_in :		in std_logic_vector(3 downto 0);
		shift_count_in :	in integer range 0 to 63;
		enable :			in std_logic;
		reset :				in std_logic;
		clock :				in std_logic;
			
		dividend_out :		out std_logic_vector(63 downto 0);
		divisor_out :		out std_logic_vector(63 downto 0);
		shift_count_out :	out integer range 0 to 63;
		resiprocal_out :	out std_logic_vector(63 downto 0);
		div_by_zero :		out std_logic);
end NR_stages;

architecture behavior of NR_stages is
	
	--Number of piupeline stages used in each Newton-Raphson stage
	constant pipe_stages_one : integer := 2;
	constant pipe_stages_two : integer := 4;
	constant pipe_stages_tree : integer := 6;
	constant pipe_stages_four : integer := 8;
	
	constant index_one : integer := 0;
	constant index_two  : integer := index_one + pipe_stages_one;
	constant index_tree  : integer := index_two + pipe_stages_two;
	constant index_four  : integer := index_tree + pipe_stages_tree;
	constant index_final  : integer := index_four + pipe_stages_four;
	
	type pipe is array (natural range <>) of std_logic_vector(63 downto 0);
	type int_pipe is array(natural range <>) of integer range 0 to 63;
	
	signal dividend_pipe : pipe(1 to index_final);
	signal divisor_pipe : pipe(1 to index_final);
	signal divisor_norm_pipe : pipe(1 to index_four);
	signal div_by_zero_pipe : std_logic_vector(index_four + 1 to index_final);
	signal shift_count_pipe : int_pipe(1 to index_final);
	
	signal resiprocal_second : std_logic_vector(7 downto 0);
	signal resiprocal_third : std_logic_vector(15 downto 0);
	signal resiprocal_fourth : std_logic_vector(31 downto 0);
	
begin

	--Pipeline bypass data
	clock_pipes:process(clock)
	begin
					
		if clock'event and clock = '1' then
			if reset = '1' then
				dividend_pipe <= (others => (others => '0'));
				
				divisor_pipe <= (others => (others => '0'));
				
				divisor_norm_pipe <= (others => (others => '0'));
				
				div_by_zero_pipe <= (others => '0');
				
				shift_count_pipe <= (others => 0);
			elsif enable = '1' then
				dividend_pipe(1) <= dividend_in;
				for i in 2 to dividend_pipe'high loop
					dividend_pipe(i) <= dividend_pipe(i-1);
				end loop;

				divisor_pipe(1) <= divisor_in;
				for i in 2 to divisor_pipe'high loop
					divisor_pipe(i) <= divisor_pipe(i-1);
				end loop;

				divisor_norm_pipe(1) <= divisor_norm_in;
				for i in 2 to divisor_norm_pipe'high loop
					divisor_norm_pipe(i) <= divisor_norm_pipe(i-1);
				end loop;

				div_by_zero_pipe(div_by_zero_pipe'low) <= not(divisor_norm_pipe(divisor_norm_pipe'high)(63));
				for i in div_by_zero_pipe'low + 1 to div_by_zero_pipe'high loop
					div_by_zero_pipe(i) <= div_by_zero_pipe(i-1);
				end loop;

				shift_count_pipe(1) <= shift_count_in;
				for i in  2 to shift_count_pipe'high loop
					shift_count_pipe(i) <= shift_count_pipe(i-1);
				end loop;
			end if;			
		end if;
	end process; 
	
	--Instantiate first Newton-Rapson stage
	NR_stage_one:NR_stage_nopipe
		generic map(4)
		port map(
			divisor_norm_in, resiprocal_in,
			enable, reset,	clock, resiprocal_second);
	
	--Instantiate second Newton-Rapson stage		
	NR_stage_two:NR_stage_nopipe
		generic map(8)
		port map(
			divisor_norm_pipe(index_two), resiprocal_second,
			enable, reset,	clock, resiprocal_third);			

	--Instantiate third Newton-Rapson stage
	NR_stage_tree:NR_stage
		generic map(16, pipe_stages_tree)
		port map(
			divisor_norm_pipe(index_tree), resiprocal_third,
			enable, reset,	clock, resiprocal_fourth);

	--Instantiate fourth Newton-Rapson stage
	NR_stage_four:NR_stage
		generic map(32, pipe_stages_four)
		port map(
			divisor_norm_pipe(index_four), resiprocal_fourth,
			enable, reset,	clock, resiprocal_out);
			
	
	dividend_out <= dividend_pipe(dividend_pipe'high); 
	divisor_out <= divisor_pipe(divisor_pipe'high);
	div_by_zero <= div_by_zero_pipe(div_by_zero_pipe'high);
	shift_count_out <= shift_count_pipe(shift_count_pipe'high);
				
end behavior;
