----------------------------------------------------------------------------------
--!@file
--!@brief First high bit position decoder and barrel shifter for 64, 32 and 16 bit opperands
-- Create Date:    16:31:50 02/20/2009 
-- Design Name: 		
-- Module Name:    barrel_shifter

-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
--!@file
--!@details This module is a combinatoric circuit that finds the highest bit that is '1' 
--!and shifts the entire word up so that this '1' gest located at the MSB
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;



entity barrel_shifter is
    generic(
		width : 			integer := 64;
		pipe_stages :	integer := 4);
	 port ( 
		input : 		in		std_logic_vector (width - 1 downto 0);
		clock :		in		std_logic;
		
		shifted : 	out	std_logic_vector (width - 1 downto 0);
		zero : 		out	std_logic;
		count : 		out	integer range 0 to width - 1);
end barrel_shifter;

--!@brief Architecture that normalizes a 64 bit integer input to the range [1, 2)
--!@return the shifted value and number of positions shifted 
architecture behavioral64 of barrel_shifter is
		
	--!pipe buffer type used for data output
	type data_pipe_t is array(0 to pipe_stages-1) of std_logic_vector(width - 1 downto 0);
	--!pipe buffer type used for shift count 
	type count_pipe_t is array(0 to pipe_stages-1) of integer range 0 to width - 1;
	
	
	--!or chain output
	signal or_output : 			std_logic_vector((64/8) - 2 downto 0);
	--!and chain output
	signal and_output :			std_logic_vector(63 downto 0);
	--!bit indicating wether or not to shift input 32 positions
	signal shift_select_32 :	std_logic;
	--!bit indicating wether or not to shift input 16 positions
	signal shift_select_16 :	std_logic;
	--!bit indicating wether or not to shift input 8 positions
	signal shift_select_8 :		std_logic;
	--!hot one coding of shift in range 0 to 7
	signal shift_select_8to0 :	std_logic_vector(7 downto 0);
	--!three step mux output
	signal three_step_mux :		std_logic_vector(63 downto 0);
	--!output data pipe
	signal data_pipe : data_pipe_t;
	--!output count pipe
	signal count_pipe : count_pipe_t;
	
begin


	--!creates chains of or'ed and and'ed input with length of 8
	and_or_chain:process(input)
		variable temp_or_output :		std_logic_vector(63 downto 0);
		variable temp_and_output :		std_logic_vector(63 downto 0);
	begin	
		for i in (64/8)-1 downto 0 loop
			temp_or_output(i*8 + 7) := input(i*8 + 7);
			temp_and_output(i*8 + 7) := input(i*8 + 7);
			for j in 6 downto 0 loop
				temp_or_output(i*8 + j) := temp_or_output((i*8)+j+1) or input((i*8)+j);
				temp_and_output(i*8 + j) := not(temp_or_output((i*8)+j+1)) and temp_or_output((i*8)+j);
			end loop;
		end loop;
		or_output <=	 temp_or_output(64-8) & temp_or_output(64-16) & temp_or_output(64-24)&
							temp_or_output(64-32) & temp_or_output(64-40) & temp_or_output(64-48)&
							temp_or_output(64-54);
		and_output <= temp_and_output;
	end process;
		
	--!generates a bit indicating whether or not to shift the input 32 positions
	shift_select_32 <= or_output(6) or or_output(5) or
								or_output(4) or or_output(3);
	
	--!generates a bit indicating whether or not to shift the input 16 positions
	shift_16:process(or_output, shift_select_32)
		variable low, high : std_logic;
	begin
		high := or_output(6) or or_output(5);
		low := or_output(2) or or_output(1);
		
		--the output is dependent on wether the high word has a '1' or not
		case shift_select_32 is
			when '1' => shift_select_16 <= high;
			when others => shift_select_16 <= low;
		end case;
	end process;

	--!generates a bit indicating whether or not to shift the input 8 positions
	shift_8:process(or_output, shift_select_32, shift_select_16)
		variable selector : std_logic_vector(1 downto 0);
	begin
		selector := shift_select_32 & shift_select_16;
		
		case selector is
			when "11" => shift_select_8 <= or_output(6);
			when "10" => shift_select_8 <= or_output(4);
			when "01" => shift_select_8 <= or_output(2);
			when others => shift_select_8 <= or_output(0);
		end case;
	end process;
	
	--!generate hot one coding for shift 0 to 8
	hot_one:process(and_output, shift_select_32, shift_select_16, shift_select_8)
		variable selector : std_logic_vector(2 downto 0);
		variable temp, temp1 :std_logic_vector(7 downto 0);
	begin
		selector := shift_select_32 & shift_select_16 & shift_select_8;
		
		case selector is
			when "000" => temp := and_output(8*1 - 1 downto 8*0);
			when "001" => temp := and_output(8*2 - 1 downto 8*1);
			when "010" => temp := and_output(8*3 - 1 downto 8*2);
			when "011" => temp := and_output(8*4 - 1 downto 8*3);
			when "100" => temp := and_output(8*5 - 1 downto 8*4);
			when "101" => temp := and_output(8*6 - 1 downto 8*5);
			when "110" => temp := and_output(8*7 - 1 downto 8*6);
			when others => temp := and_output(8*8 - 1 downto 8*7);
		end case;
		
		--swap range
		for i in temp'range loop
			temp1(7-i) := temp(i);
		end loop;
		
		shift_select_8to0 <= temp1;
	end process;
	

	--!three step mux for shifing input 32, 16 and 8 positions 
	big_shift:process(input, shift_select_32, shift_select_16, shift_select_8)
		variable step_one :		std_logic_vector(63 downto 0);
		variable step_two :		std_logic_vector(63 downto 0);
		variable step_three :		std_logic_vector(63 downto 0);
	begin
		
		--first mux step, shift 32 bit left
		if shift_select_32 = '1' then
			step_one := input;
		else
			step_one(63 downto 32) := input(63 - 32 downto 0);
			step_one(31 downto 0) := (others => '0');
		end if;
		
		--second mux step, shift 16 bit left
		if shift_select_16 = '1' then
			step_two := step_one;
		else
			step_two(63 downto 16) := step_one(63 - 16 downto 0);
			step_two(15 downto 0) := (others => '0');
		end if; 
		
		--third mux step, shift 8 bit left
		if shift_select_8 = '1' then
			step_three := step_two;
		else
			step_three(63 downto 8) := step_two(63 - 8 downto 0);
			step_three(7 downto 0) := (others => '0');
		end if;
		
		three_step_mux <= step_three;
		
	end process;
	
	--!the last 0 to 8 shift is done using hardware multipliers to achive greater speed
	--!@note pipelined
	mult_shift:process(clock)
		variable shift_out : std_logic_vector(three_step_mux'length + shift_select_8to0'length -1 downto 0);
	begin
		if clock'event and clock = '1' then
			--temp_out <= pipe(3);--trucate to the 64 least significant bits
		
			for i in data_pipe'high downto 1 loop
				data_pipe(i) <= data_pipe(i-1);
			end loop;
			
			shift_out := (three_step_mux * shift_select_8to0);
			data_pipe(0) <= shift_out(data_pipe(0)'range);
		end if;
	end process;
	
	--!assign output value
	shifted <= data_pipe(data_pipe'high);
	
	--!if the most significant bit is zero after the shift it means that the input was zero
	zero <= not data_pipe(data_pipe'high)(63);
	
	--Code the shift count from hot one coding to in binary format
	shift_count_coding:process(clock)
		variable binary : std_logic_vector(2 downto 0);
	begin
		if clock'event and clock = '1' then
			--!implements 8 to 3 coder
			if 	shift_select_8to0 = "00000001" then binary := "000";
			elsif shift_select_8to0 = "00000010" then binary := "001";
			elsif shift_select_8to0 = "00000100" then binary := "010";
			elsif shift_select_8to0 = "00001000" then binary := "011";
			elsif shift_select_8to0 = "00010000" then binary := "100";
			elsif shift_select_8to0 = "00100000" then binary := "101";
			elsif shift_select_8to0 = "01000000" then binary := "110";
			elsif shift_select_8to0 = "10000000" then binary := "111";
			else binary := "000"; end if;
			
			--it should never hapen that shift_select_8to0 is not hot one coded
			--assert (shift_select_8to0 /= "00000001" and binary = "000")
			--report "invalid shift_select_8to0"
			--severity failure;
			
			
			for i in count_pipe'high downto 1 loop
				count_pipe(i) <= count_pipe(i-1);
			end loop;
			
			count_pipe(0) <= conv_integer(not(shift_select_32 & shift_select_16 & shift_select_8) & binary);
		end if;
		
	end process;
	
	count <= count_pipe(count_pipe'high);
	
	
end behavioral64;
