------------------------------------------------------
--(copy left) Arthur Øvergaard 
--Trondheim June 09
--
--!@file
--!@details This module is a combinatoric circuit that finds the highest bit that is '1' 
--!and shifts the entire word up so that this '1' gest located at the MSB
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;



entity norm_shifter is
    generic(pipe_stages :	integer := 4);
	 port ( 
		dividend_in :	in std_logic_vector(63 downto 0);
		divisor_in :	in std_logic_vector(63 downto 0);
		enable :		in std_logic;
		reset : 		in std_logic;
		clock :			in std_logic;
		
		divisor_norm : 	out	std_logic_vector(63 downto 0);
		dividend_out :	out std_logic_vector(63 downto 0);
		divisor_out :	out std_logic_vector(63 downto 0);
		count : 		out	integer range 0 to 63);
end norm_shifter;

--!@brief Architecture that normalizes a 64 bit integer input to the range [1, 2)
--!@return the shifted value and number of positions shifted 
architecture behavior of norm_shifter is
		
	--!pipe buffer type used for data output
	type pipe is array (natural range <>) of std_logic_vector(63 downto 0);
	--!pipe buffer type used for shift count 
	type count_pipe_t is array (natural range <>) of integer range 0 to 63;
	
	
	--!or chain output
	signal or_output : 			std_logic_vector((64/8) - 2 downto 0);
	--!and chain output
	signal and_output :			std_logic_vector(63 downto 0);
	--!bit indicating wether or not to shift input 32 positions
	signal shift_select_32 :	std_logic;
	--!bit indicating wether or not to shift input 16 positions
	signal shift_select_16 :	std_logic;
	--!bit indicating wether or not to shift input 8 positions
	signal shift_select_8 :		std_logic;
	--!hot one coding of shift in range 0 to 7
	signal shift_select_8to0 :	std_logic_vector(7 downto 0);
	--!three step mux output
	signal three_step_mux :		std_logic_vector(63 downto 0);
	--!output data pipe
	signal data_pipe : pipe(0 to pipe_stages - 1);
	--!output count pipe
	signal count_pipe : count_pipe_t(0 to pipe_stages - 1);
	
	signal dividend_pipe : pipe(0 to pipe_stages - 1);
	signal divisor_pipe : pipe(0 to pipe_stages - 1);
	
begin


	--!creates chains of or'ed and and'ed input with length of 8
	and_or_chain:process(divisor_in)
		variable temp_or_output :		std_logic_vector(63 downto 0);
		variable temp_and_output :		std_logic_vector(63 downto 0);
	begin	
		for i in (64/8)-1 downto 0 loop
			temp_or_output(i*8 + 7) := divisor_in(i*8 + 7);
			temp_and_output(i*8 + 7) := divisor_in(i*8 + 7);
			for j in 6 downto 0 loop
				temp_or_output(i*8 + j) := temp_or_output((i*8)+j+1) or divisor_in((i*8)+j);
				temp_and_output(i*8 + j) := not(temp_or_output((i*8)+j+1)) and temp_or_output((i*8)+j);
			end loop;
		end loop;
		or_output <=	 temp_or_output(64-8) & temp_or_output(64-16) & temp_or_output(64-24)&
							temp_or_output(64-32) & temp_or_output(64-40) & temp_or_output(64-48)&
							temp_or_output(64-56);
		and_output <= temp_and_output;
	end process;
		
	--!generates a bit indicating whether or not to shift the input 32 positions
	shift_select_32 <= or_output(6) or or_output(5) or
								or_output(4) or or_output(3);
	
	--!generates a bit indicating whether or not to shift the input 16 positions
	shift_16:process(or_output, shift_select_32)
		variable low, high : std_logic;
	begin
		high := or_output(6) or or_output(5);
		low := or_output(2) or or_output(1);
		
		--the output is dependent on wether the high word has a '1' or not
		case shift_select_32 is
			when '1' => shift_select_16 <= high;
			when others => shift_select_16 <= low;
		end case;
	end process;

	--!generates a bit indicating whether or not to shift the input 8 positions
	shift_8:process(or_output, shift_select_32, shift_select_16)
		variable selector : std_logic_vector(1 downto 0);
	begin
		selector := shift_select_32 & shift_select_16;
		
		case selector is
			when "11" => shift_select_8 <= or_output(6);
			when "10" => shift_select_8 <= or_output(4);
			when "01" => shift_select_8 <= or_output(2);
			when others => shift_select_8 <= or_output(0);
		end case;
	end process;
	
	--!generate hot one coding for shift 0 to 8
	hot_one:process(and_output, shift_select_32, shift_select_16, shift_select_8)
		variable selector : std_logic_vector(2 downto 0);
		variable temp, temp1 :std_logic_vector(7 downto 0);
	begin
		selector := shift_select_32 & shift_select_16 & shift_select_8;
		
		case selector is
			when "000" => temp := and_output(8*1 - 1 downto 8*0);
			when "001" => temp := and_output(8*2 - 1 downto 8*1);
			when "010" => temp := and_output(8*3 - 1 downto 8*2);
			when "011" => temp := and_output(8*4 - 1 downto 8*3);
			when "100" => temp := and_output(8*5 - 1 downto 8*4);
			when "101" => temp := and_output(8*6 - 1 downto 8*5);
			when "110" => temp := and_output(8*7 - 1 downto 8*6);
			when others => temp := and_output(8*8 - 1 downto 8*7);
		end case;
		
		--swap range
		for i in temp'range loop
			temp1(7-i) := temp(i);
		end loop;
		
		shift_select_8to0 <= temp1;
	end process;
	

	--!three step mux for shifing input 32, 16 and 8 positions 
	big_shift:process(divisor_in, shift_select_32, shift_select_16, shift_select_8)
		variable step_one :		std_logic_vector(63 downto 0);
		variable step_two :		std_logic_vector(63 downto 0);
		variable step_three :	std_logic_vector(63 downto 0);
	begin
		
		--first mux step, shift 32 bit left
		if shift_select_32 = '1' then
			step_one := divisor_in;
		else
			step_one(63 downto 32) := divisor_in(63 - 32 downto 0);
			step_one(31 downto 0) := (others => '0');
		end if;
		
		--second mux step, shift 16 bit left
		if shift_select_16 = '1' then
			step_two := step_one;
		else
			step_two(63 downto 16) := step_one(63 - 16 downto 0);
			step_two(15 downto 0) := (others => '0');
		end if; 
		
		--third mux step, shift 8 bit left
		if shift_select_8 = '1' then
			step_three := step_two;
		else
			step_three(63 downto 8) := step_two(63 - 8 downto 0);
			step_three(7 downto 0) := (others => '0');
		end if;
		
		three_step_mux <= step_three;
		
	end process;
	
	--!the last 0 to 8 shift is done using hardware multipliers to achive greater speed
	--!@note pipelined
	mult_shift:process(clock)
		variable shift_out : std_logic_vector(three_step_mux'length + shift_select_8to0'length -1 downto 0);
	begin

		if clock'event and clock = '1' then
			if reset = '1' then
				for i in 0 to data_pipe'high loop
					data_pipe(i) <= (others => '0');
				end loop;		
			elsif enable = '1' then
				--temp_out <= pipe(3);--trucate to the 64 least significant bits
			
				shift_out := (three_step_mux * shift_select_8to0);
				data_pipe(0) <= shift_out(data_pipe(0)'range);
			
				for i in 1 to data_pipe'high loop
					data_pipe(i) <= data_pipe(i-1);
				end loop;
			end if;
		end if;
	end process;
	
	
	
	shift_count_coding:process(clock)
		variable binary : std_logic_vector(2 downto 0);
	begin
		
		if clock'event and clock = '1' then
			if reset = '1' then
				for i in 0 to count_pipe'high loop
					count_pipe(i) <= 0;
				end loop;		
			elsif enable = '1' then
				--!implements 8 to 3 coder
				if 	shift_select_8to0 = "00000001" then binary := "000";
				elsif shift_select_8to0 = "00000010" then binary := "001";
				elsif shift_select_8to0 = "00000100" then binary := "010";
				elsif shift_select_8to0 = "00001000" then binary := "011";
				elsif shift_select_8to0 = "00010000" then binary := "100";
				elsif shift_select_8to0 = "00100000" then binary := "101";
				elsif shift_select_8to0 = "01000000" then binary := "110";
				elsif shift_select_8to0 = "10000000" then binary := "111";
				else binary := "000"; end if;
			
				--it should never hapen that shift_select_8to0 is not hot one coded
				--assert (shift_select_8to0 /= "00000001" and binary = "000")
				--report "invalid shift_select_8to0"
				--severity warning;
			
				count_pipe(0) <= conv_integer(not(shift_select_32 & shift_select_16 & shift_select_8) & binary);
			
				for i in 1 to count_pipe'high loop
					count_pipe(i) <= count_pipe(i-1);
				end loop;
			end if;	
		end if;
		
	end process;
	
	pipe_throug:process(clock)
	begin
		if clock'event and clock='1' then
			if reset = '1' then
				for i in 0 to dividend_pipe'high loop
					dividend_pipe(i) <= (others => '0');
					divisor_pipe(i) <= (others => '0');
				end loop;		
			elsif enable = '1' then
				dividend_pipe(0) <= dividend_in;
				divisor_pipe(0) <= divisor_in;		
			
				for i in 1 to dividend_pipe'high loop
					dividend_pipe(i) <= dividend_pipe(i-1);
					divisor_pipe(i) <= divisor_pipe(i-1);
				end loop;
			end if;
		end if;
	end process;
	
	
	divisor_norm <= data_pipe(data_pipe'high);
	count <= count_pipe(count_pipe'high);
	dividend_out <= dividend_pipe(dividend_pipe'high);
	divisor_out <= divisor_pipe(divisor_pipe'high);
	
end behavior;
