------------------------------------------------------
--(copy left) Arthur Øvergaard 
--Trondheim June 09
--
--Implements the quotient calculation stage
------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
use IEEE.numeric_std.all;


entity quotient_calc is
	generic(pipe_stages1 : integer := 1;
			pipe_stages2 : integer := 1);
	port(
		resciprocal :	in std_logic_vector(63 downto 0);
		dividend_in :	in std_logic_vector(63 downto 0);
		divisor_in :	in std_logic_vector(63 downto 0);
		div_by_zero_in :in std_logic;
		shift_count :	in integer range 63 downto 0;
		enable :		in std_logic;
		reset :			in std_logic;
		clock :			in std_logic;
		
		dividend_out :	out std_logic_vector(63 downto 0);
		divisor_out :	out std_logic_vector(63 downto 0);
		div_by_zero_out :out std_logic;
		quotient : 		out std_logic_vector(63 downto 0));
end quotient_calc;

architecture behavior of quotient_calc is
	constant n : integer := 64;
	
	type mult_pipe_t is array (0 to pipe_stages1 - 1) of std_logic_vector((2 * n) - 1 downto 0);
	type shift_pipe_t is array (mult_pipe_t'range) of integer range n - 1 downto 0;
	type shift_mult_pipe_t is array (0 to pipe_stages2 - 1) of std_logic_vector(n + 8 - 1 downto 0);
	
	type pipe is array (0 to pipe_stages1 + pipe_stages2 - 1) of std_logic_vector(63 downto 0);
	
	signal dividend_pipe : pipe;
	signal divisor_pipe : pipe;
	signal zero_pipe : std_logic_vector(0 to pipe_stages1 + pipe_stages2 - 1);
	
	signal mult_pipe : mult_pipe_t;
	signal shift_pipe : shift_pipe_t;
	signal shift_mult_pipe : shift_mult_pipe_t;
	
	signal mult : std_logic_vector(n - 1 downto 0);
	signal shift_high : std_logic_vector(2 downto 0);--binary coded
	signal shift_low : std_logic_vector(7 downto 0);--hot one coded
begin

	--Pipeline all bypass data
	clock_pipe:process(clock)
	begin

		if clock'event and clock = '1' then
			if reset = '1' then
		
				dividend_pipe <= (others => (others => '0'));
				
				divisor_pipe <= (others => (others => '0'));

				zero_pipe <=(others => '0');
		
			elsif enable = '1' then
				dividend_pipe(0) <= dividend_in;
				for i in 1 to dividend_pipe'high loop
					dividend_pipe(i) <= dividend_pipe(i-1);
				end loop;
			
				divisor_pipe(0) <= divisor_in;
				for i in 1 to divisor_pipe'high loop
					divisor_pipe(i) <= divisor_pipe(i-1);
				end loop;			

				zero_pipe(0) <= div_by_zero_in;
				for i in 1 to zero_pipe'high loop
					zero_pipe(i) <= zero_pipe(i-1);
				end loop;
			end if;
		end if;
	end process;
		
	dividend_out <= dividend_pipe(dividend_pipe'high);	
	divisor_out <= divisor_pipe(divisor_pipe'high);
	div_by_zero_out <= zero_pipe(zero_pipe'high);

	--Implements the multiplication between the dividend and the reciprocal
	multiply:process(clock)
	begin
			
		if clock'event and clock = '1' then
			if reset = '1' then
				mult_pipe <= (others => (others => '0'));
			elsif enable = '1' then
		
				mult_pipe(0) <= resciprocal * dividend_in;
			
				for i in 1 to mult_pipe'high loop
					mult_pipe(i) <= mult_pipe(i-1);
				end loop;	
			end if;
		end if;
	end process;
	
	--Truncate the quotient
	mult <= mult_pipe(mult_pipe'high)((2 * n) - 1 downto n);
	
	--Pipeline the shift count data
	hold_shift_count:process(clock)
	begin
		
		if clock'event and clock = '1' then
			if reset = '1' then
				shift_pipe <= (others => 0);
			elsif enable = '1' then
				shift_pipe(0) <= shift_count;
			
				for i in 1 to shift_pipe'high loop
					shift_pipe(i) <= shift_pipe(i-1);
				end loop;				
			end if;
		end if;
	end process;
	
	--re-codes the binary coded shift count to "hot one" coding
	shift_code:process(shift_pipe(shift_pipe'high))
		variable temp : std_logic_vector(5 downto 0);
	begin
		temp := conv_std_logic_vector(shift_pipe(shift_pipe'high), temp'length);
		
		shift_high <= temp(temp'high downto temp'length - shift_high'length);
		
		case temp(2 downto 0) is
			when "000" =>  shift_low <= "00000001";
			when "001" =>  shift_low <= "00000010";
			when "010" =>  shift_low <= "00000100";
			when "011" =>  shift_low <= "00001000";
			when "100" =>  shift_low <= "00010000";
			when "101" =>  shift_low <= "00100000";
			when "110" =>  shift_low <= "01000000";
			when others => shift_low <= "10000000";
		end case;
	end process;
	
	--Implements three stage multiplixer barrel shifter and a final pipelined multiplier shifter
	right_shift:process(clock)	
		variable shift : std_logic_vector(n - 1 downto 0) := (others => '0');	
	begin	
		if clock'event and clock = '1' then
			if reset = '1' then
				shift_mult_pipe(i) <= (others=>(others => '0'));
			elsif enable = '1' then
				shift := mult;
			
				if shift_high(2) = '0' then
					shift(63 - 32 downto 0) := mult(63 downto 32);
					shift(63 downto 64 - 32) := (others => '0');
				else
					shift := mult;
				end if;
			
				if shift_high(1) = '0' then
					shift(63 - 16 downto 0) := shift(63 downto 16);
					shift(63 downto 64 - 16) := (others => '0');
				end if;

				if shift_high(0) = '0' then
					shift(63 - 8 downto 0) := shift(63 downto 8);
					shift(63 downto 64 - 8) := (others => '0');
				end if;
			
				shift_mult_pipe(0) <= shift * shift_low;
				for i in 1 to shift_mult_pipe'high loop
					shift_mult_pipe(i) <= shift_mult_pipe(i-1);
				end loop;
			end if;
		end if;
	end process;
	
	--Truncate the final output
	quotient <= shift_mult_pipe(shift_mult_pipe'high)(n + 8 - 2 downto 8-1);	
	
	
	
end behavior;
		
